local babycode = {}

---renders babycode to html
---@param s string input babycode
---@param escape_html fun(s: string): string function that escapes html
function babycode.to_html(s, escape_html)
  if not s or s == "" then return "" end
  -- extract code blocks first and store them as placeholders
  -- don't want to process bbcode embedded into a code block
  local code_blocks = {}
  local code_count = 0
  local text = s:gsub("%[code%](.-)%[/code%]", function(code)
    code_count = code_count + 1
    -- strip leading and trailing newlines, preserve others
    code_blocks[code_count] = code:gsub("^%s*(.-)%s*$", "%1")
    return "\1CODE:"..code_count.."\1"
  end)

  -- replace `[url=https://example.com]Example[/url] tags
  text = text:gsub("%[url=([^%]]+)%](.-)%[/url%]", function(url, label)
    return '<a href="'..escape_html(url)..'">'..escape_html(label)..'</a>'
  end)
  
  -- replace `[url]https://example.com[/url] tags
  text = text:gsub("%[url%]([^%]]+)%[/url%]", function(url)
    return '<a href="'..escape_html(url)..'">'..escape_html(url)..'</a>'
  end)

  -- bold, italics, strikethrough
  text = text:gsub("%[b%](.-)%[/b%]", "<strong>%1</strong>")
  text = text:gsub("%[i%](.-)%[/i%]", "<em>%1</em>")
  text = text:gsub("%[s%](.-)%[/s%]", "<del>%1</del>")

  -- replace loose links
  text = text:gsub("(https?://[%w-_%.%?%.:/%+=&~%@#%%]+[%w-/])", function(url)
    if not text:find('<a[^>]*>'..url..'</a>') then
      return '<a href="'..escape_html(url)..'">'..escape_html(url)..'</a>'
    end
    return url
  end)

  -- normalize newlines, replace them with <br>
  text = text:gsub("\r?\n\r?\n+", "<br>"):gsub("\r?\n", "<br>")

  -- replace code block placeholders back with their original contents
  text = text:gsub("\1CODE:(%d+)\1", function(n)
    return "<pre><code>"..code_blocks[tonumber(n)].."</code></pre>"
  end)

  return text
end

return babycode
